@@ -4,17 +4,31 @@ module HLint.Default where
 import Control.Arrow
 import Control.Exception
 import Control.Monad
+import Control.Monad.Trans.State
+import qualified Data.Foldable
+import Data.Foldable(asum, sequenceA_, traverse_, for_)
+import Data.Traversable(traverse, for)
+import Control.Applicative
 import Data.Function
 import Data.Int
+import Data.Char
+import Data.List as Data.List
 import Data.List as X
 import Data.Maybe
 import Data.Monoid
 import System.IO
+import Control.Concurrent.Chan
+import System.Mem.Weak
+import Control.Exception.Base
+import System.Exit
+import Data.Either
+import Numeric
 
 import IO as System.IO
 import List as Data.List
 import Maybe as Data.Maybe
 import Monad as Control.Monad
+import Char as Data.Char
 
 -- I/O
 
@@ -47,10 +61,19 @@ error = compare x y /= GT ==> x <= y
 error = compare x y == LT ==> x < y
 error = compare x y /= LT ==> x >= y
 error = compare x y == GT ==> x > y
-error = x == a || x == b || x == c ==> x `elem` [a,b,c]
-error = x /= a && x /= b && x /= c ==> x `notElem` [a,b,c]
+error = x == a || x == b || x == c ==> x `elem` [a,b,c] where note = ValidInstance "Eq" x
+error = x /= a && x /= b && x /= c ==> x `notElem` [a,b,c] where note = ValidInstance "Eq" x
 --error = compare (f x) (f y) ==> Data.Ord.comparing f x y -- not that great
 --error = on compare f ==> Data.Ord.comparing f -- not that great
+error = head (sort x) ==> minimum x
+error = last (sort x) ==> maximum x
+error = head (sortBy f x) ==> minimumBy f x
+error = last (sortBy f x) ==> maximumBy f x
+error "Avoid reverse" = reverse (sort x) ==> sortBy (flip compare) x
+error "Avoid reverse" = reverse (sortBy f x) ==> sortBy (flip f) x
+warn  = flip (g `on` h) ==> flip g `on` h
+warn  = (f `on` g) `on` h ==> f `on` (g . h)
+
 
 -- READ/SHOW
 
@@ -63,27 +86,38 @@ warn = showIntAtBase 8 intToDigit ==> showOct
 -- LIST
 
 error = concat (map f x) ==> concatMap f x
-warn = concat [a,b] ==> a ++ b
+warn = concat [a, b] ==> a ++ b
 warn "Use map once" = map f (map g x) ==> map (f . g) x
 warn  = x !! 0 ==> head x
 error = take n (repeat x) ==> replicate n x
+error = map f (replicate n x) ==> replicate n (f x)
+error = map f (repeat x) ==> repeat (f x)
 error = head (reverse x) ==> last x
-error = head (drop n x) ==> x !! n
-error = reverse (tail (reverse x)) ==> init x
-error = take (length x - 1) x ==> init x
+error = head (drop n x) ==> x !! n where _ = isNat n
+error = reverse (tail (reverse x)) ==> init x where note = IncreasesLaziness
+error "Avoid reverse" = reverse (reverse x) ==> x where note = IncreasesLaziness
+-- error = take (length x - 1) x ==> init x -- not true for x == []
 error = isPrefixOf (reverse x) (reverse y) ==> isSuffixOf x y
 error = foldr (++) [] ==> concat
-error = foldl (++) [] ==> concat
+error = foldl (++) [] ==> concat where note = IncreasesLaziness
 error = span (not . p) ==> break p
 error = break (not . p) ==> span p
+error = (takeWhile p x, dropWhile p x) ==> span p x
+error = fst (span p x) ==> takeWhile p x
+error = snd (span p x) ==> dropWhile p x
+error = fst (break p x) ==> takeWhile (not . p) x
+error = snd (break p x) ==> dropWhile (not . p) x
 error = concatMap (++ "\n") ==> unlines
+error = intercalate "\n" x ++ "\n" ==> unlines x
+error = map id ==> id
 error = or (map p x) ==> any p x
 error = and (map p x) ==> all p x
 error = zipWith (,) ==> zip
 error = zipWith3 (,,) ==> zip3
-warn  = length x == 0 ==> null x where note = "increases laziness"
-warn  "Use null" = length x /= 0 ==> not (null x)
-error "Use :" = (\x -> [x]) ==> (:[])
+warn  = length x == 0 ==> null x where note = IncreasesLaziness
+warn  = x == [] ==> null x
+warn  "Use null" = length x /= 0 ==> not (null x) where note = IncreasesLaziness
+warn  "Use :" = (\x -> [x]) ==> (:[])
 error = map (uncurry f) (zip x y) ==> zipWith f x y
 warn  = map f (zip x y) ==> zipWith (curry f) x y where _ = isVar f
 error = not (elem x y) ==> notElem x y
@@ -95,32 +129,49 @@ warn  = concat (intersperse " " x) ==> unwords x
 error "Use any" = null (filter f x) ==> not (any f x)
 error "Use any" = filter f x == [] ==> not (any f x)
 error = filter f x /= [] ==> any f x
-error = any ((==) a) ==> elem a
+error = any id ==> or
+error = all id ==> and
+error = any ((==) a) ==> elem a where note = ValidInstance "Eq" a
 error = any (== a) ==> elem a
-error = any (a ==) ==> elem a
-error = all ((/=) a) ==> notElem a
-error = all (/= a) ==> notElem a
-error = all (a /=) ==> notElem a
+error = any (a ==) ==> elem a where note = ValidInstance "Eq" a
+error = all ((/=) a) ==> notElem a where note = ValidInstance "Eq" a
+error = all (/= a) ==> notElem a where note = ValidInstance "Eq" a
+error = all (a /=) ==> notElem a where note = ValidInstance "Eq" a
+error = elem True ==> or
+error = notElem False ==> and
+error = findIndex ((==) a) ==> elemIndex a
+error = findIndex (a ==) ==> elemIndex a
+error = findIndex (== a) ==> elemIndex a
+error = findIndices ((==) a) ==> elemIndices a
+error = findIndices (a ==) ==> elemIndices a
+error = findIndices (== a) ==> elemIndices a
+error = lookup b (zip l [0..]) ==> elemIndex b l
+warn "Length always non-negative" = length x >= 0 ==> True
+warn "Use null" = length x > 0 ==> not (null x) where note = IncreasesLaziness
+warn "Use null" = length x >= 1 ==> not (null x) where note = IncreasesLaziness
+error "Take on a non-positive" = take i x ==> [] where _ = isNegZero i
+error "Drop on a non-positive" = drop i x ==> x where _ = isNegZero i
+
 
 -- FOLDS
 
 error = foldr  (>>) (return ()) ==> sequence_
 error = foldr  (&&) True ==> and
-error = foldl  (&&) True ==> and
-error = foldr1 (&&)  ==> and
-error = foldl1 (&&)  ==> and
+error = foldl  (&&) True ==> and where note = IncreasesLaziness
+error = foldr1 (&&)  ==> and where note = RemovesError "on []"
+error = foldl1 (&&)  ==> and where note = RemovesError "on []"
 error = foldr  (||) False ==> or
-error = foldl  (||) False ==> or
-error = foldr1 (||)  ==> or
-error = foldl1 (||)  ==> or
+error = foldl  (||) False ==> or where note = IncreasesLaziness
+error = foldr1 (||)  ==> or where note = RemovesError "on []"
+error = foldl1 (||)  ==> or where note = RemovesError "on []"
 error = foldl  (+) 0 ==> sum
 error = foldr  (+) 0 ==> sum
-error = foldl1 (+)   ==> sum
-error = foldr1 (+)   ==> sum
+error = foldl1 (+)   ==> sum where note = RemovesError "on []"
+error = foldr1 (+)   ==> sum where note = RemovesError "on []"
 error = foldl  (*) 1 ==> product
 error = foldr  (*) 1 ==> product
-error = foldl1 (*)   ==> product
-error = foldr1 (*)   ==> product
+error = foldl1 (*)   ==> product where note = RemovesError "on []"
+error = foldr1 (*)   ==> product where note = RemovesError "on []"
 error = foldl1 max   ==> maximum
 error = foldr1 max   ==> maximum
 error = foldl1 min   ==> minimum
@@ -131,15 +182,18 @@ error = foldr mplus mzero ==> msum
 
 error = (\x -> x) ==> id
 error = (\x y -> x) ==> const
-error = (\(_,y) -> y) ==> snd
-error = (\(x,_) -> x) ==> fst
-warn "Use curry" = (\x y-> f (x,y)) ==> curry f where _ = notIn [x,y] f
-warn "Use uncurry" = (\(x,y) -> f x y) ==> uncurry f where _ = notIn [x,y] f
+error = (\(x,y) -> y) ==> snd
+error = (\(x,y) -> x) ==> fst
+warn "Use curry" = (\x y -> f (x,y)) ==> curry f
+warn "Use uncurry" = (\(x,y) -> f x y) ==> uncurry f where note = IncreasesLaziness
 error "Redundant $" = (($) . f) ==> f
 error "Redundant $" = (f $) ==> f
 warn  = (\x -> y) ==> const y where _ = isAtom y && notIn x y
 error "Redundant flip" = flip f x y ==> f y x where _ = isApp original
-warn  = (\a b -> o (f a) (f b)) ==> o `Data.Function.on` f
+warn  = (\a b -> g (f a) (f b)) ==> g `Data.Function.on` f
+error "Evaluate" = id x ==> x
+error "Redundant id" = id . x ==> x
+error "Redundant id" = x . id ==> x
 
 -- CHAR
 
@@ -149,13 +203,21 @@ error = a >= '0' && a <= '9' ==> isDigit a
 error = a >= '0' && a <= '7' ==> isOctDigit a
 error = not (isControl a) ==> isPrint a
 error = isLower a || isUpper a ==> isAlpha a
+error = isUpper a || isLower a ==> isAlpha a
 error = isAlpha a || isDigit a ==> isAlphaNum a
+error = isDigit a || isAlpha a ==> isAlphaNum a
 
 -- BOOL
 
-error "Redundant ==" = a == True ==> a
-warn  "Redundant ==" = a == False ==> not a
-error "Redundant if" = (if a then x else x) ==> x where note = "reduces strictness"
+error "Redundant ==" = x == True ==> x
+warn  "Redundant ==" = x == False ==> not x
+error "Redundant ==" = True == a ==> a
+warn  "Redundant ==" = False == a ==> not a
+error "Redundant /=" = a /= True ==> not a
+warn  "Redundant /=" = a /= False ==> a
+error "Redundant /=" = True /= a ==> not a
+warn  "Redundant /=" = False /= a ==> a
+error "Redundant if" = (if a then x else x) ==> x where note = IncreasesLaziness
 error "Redundant if" = (if a then True else False) ==> a
 error "Redundant if" = (if a then False else True) ==> not a
 error "Redundant if" = (if a then t else (if b then t else f)) ==> if a || b then t else f
@@ -166,29 +228,35 @@ warn  "Use if" = case a of {True -> t; False -> f} ==> if a then t else f
 warn  "Use if" = case a of {False -> f; True -> t} ==> if a then t else f
 warn  "Use if" = case a of {True -> t; _ -> f} ==> if a then t else f
 warn  "Use if" = case a of {False -> f; _ -> t} ==> if a then t else f
-warn  "Redundant if" = (if c then (True, x) else (False, x)) ==> (c, x) where note = "reduces strictness"
-warn  "Redundant if" = (if c then (False, x) else (True, x)) ==> (not c, x) where note = "reduces strictness"
-warn = or [x,y]  ==> x || y
-warn = or [x,y,z]  ==> x || y || z
-warn = and [x,y]  ==> x && y
-warn = and [x,y,z]  ==> x && y && z
+warn  "Redundant if" = (if c then (True, x) else (False, x)) ==> (c, x) where note = IncreasesLaziness
+warn  "Redundant if" = (if c then (False, x) else (True, x)) ==> (not c, x) where note = IncreasesLaziness
+warn = or [x, y] ==> x || y
+warn = or [x, y, z] ==> x || y || z
+warn = and [x, y] ==> x && y
+warn = and [x, y, z] ==> x && y && z
+error "Redundant if" = (if x then False else y) ==> not x && y where _ = notEq y True
+error "Redundant if" = (if x then y else True) ==> not x || y where _ = notEq y False
+error "Redundant not" = not (not x) ==> x
+error "Too strict if" = (if c then f x else f y) ==> f (if c then x else y) where note = IncreasesLaziness
 
 -- ARROW
 
 error = id *** g ==> second g
 error = f *** id ==> first f
 error = zip (map f x) (map g x) ==> map (f Control.Arrow.&&& g) x
-warn  = (\(x,y) -> (f x, g y)) ==> f Control.Arrow.*** g where _ = notIn [x,y] [f,g]
-warn  = (\x -> (f x, g x)) ==> f Control.Arrow.&&& g where _ = notIn x [f,g]
-warn  = (\(x,y) -> (f x,y)) ==> Control.Arrow.first f where _ = notIn [x,y] f
-warn  = (\(x,y) -> (x,f y)) ==> Control.Arrow.second f where _ = notIn [x,y] f
+warn  = (\(x,y) -> (f x, g y)) ==> f Control.Arrow.*** g
+warn  = (\x -> (f x, g x)) ==> f Control.Arrow.&&& g
+warn  = (\(x,y) -> (f x,y)) ==> Control.Arrow.first f
+warn  = (\(x,y) -> (x,f y)) ==> Control.Arrow.second f
 warn  = (f (fst x), g (snd x)) ==> (f Control.Arrow.*** g) x
-warn "Redundant pair" = (fst x, snd x) ==>  x
+warn "Redundant pair" = (fst x, snd x) ==>  x where note = DecreasesLaziness
 
 -- FUNCTOR
 
 error "Functor law" = fmap f (fmap g x) ==> fmap (f . g) x
 error "Functor law" = fmap id ==> id
+warn = fmap f $ x ==> f Control.Applicative.<$> x
+    where _ = isApp x || isAtom x
 
 -- MONAD
 
@@ -208,16 +276,23 @@ warn  = flip forM_ ==> mapM_
 error = when (not x) ==> unless x
 error = x >>= id ==> Control.Monad.join x
 error = liftM f (liftM g x) ==> liftM (f . g) x
-warn = a >> return () ==> void a
-warn = fmap (const ()) ==> void
+error = fmap f (fmap g x) ==> fmap (f . g) x
+warn  = a >> return () ==> Control.Monad.void a
+    where _ = isAtom a || isApp a
+error = fmap (const ()) ==> Control.Monad.void
 error = flip (>=>) ==> (<=<)
 error = flip (<=<) ==> (>=>)
-error = (\x -> f x >>= g) ==> f Control.Monad.>=> g where _ = notIn x [f,g]
-error = (\x -> f =<< g x) ==> f Control.Monad.<=< g where _ = notIn x [f,g]
+warn  = (\x -> f x >>= g) ==> f Control.Monad.>=> g
+warn  = (\x -> f =<< g x) ==> f Control.Monad.<=< g
 error = a >> forever a ==> forever a
-warn = liftM2 id ==> ap
+warn  = liftM2 id ==> ap
 error = mapM (uncurry f) (zip l m) ==> zipWithM f l m
 
+-- STATE MONAD
+
+error = fst (runState x y) ==> evalState x y
+error = snd (runState x y) ==> execState x y
+
 -- MONAD LIST
 
 error = liftM unzip (mapM f x) ==> Control.Monad.mapAndUnzipM f x
@@ -225,8 +300,12 @@ error = sequence (zipWith f x y) ==> Control.Monad.zipWithM f x y
 error = sequence_ (zipWith f x y) ==> Control.Monad.zipWithM_ f x y
 error = sequence (replicate n x) ==> Control.Monad.replicateM n x
 error = sequence_ (replicate n x) ==> Control.Monad.replicateM_ n x
+error = mapM f (replicate n x) ==> Control.Monad.replicateM n (f x)
+error = mapM_ f (replicate n x) ==> Control.Monad.replicateM_ n (f x)
 error = mapM f (map g x) ==> mapM (f . g) x
 error = mapM_ f (map g x) ==> mapM_ (f . g) x
+error = mapM id ==> sequence
+error = mapM_ id ==> sequence_
 
 -- APPLICATIVE / TRAVERSABLE
 
@@ -280,6 +359,9 @@ error = isJust x && (fromJust x == y) ==> x == Just y
 error = mapMaybe f (map g x) ==> mapMaybe (f . g) x
 error = fromMaybe a (fmap f x) ==> maybe a f x
 warn = [x | Just x <- a] ==> Data.Maybe.catMaybes a
+warn = (case m of Nothing -> Nothing; Just x -> x) ==> Control.Monad.join m
+warn = maybe Nothing id ==> join
+warn "Too strict maybe" = maybe (f x) (f . g) ==> f . maybe x g where note = IncreasesLaziness
 
 -- EITHER
 
@@ -303,7 +385,6 @@ error "Redundant fromInteger" = fromInteger x ==> x where _ = isLitInt x
 warn  = x + negate y ==> x - y
 warn  = 0 - x ==> negate x
 warn  = log y / log x ==> logBase x y
-warn  = x ** 0.5 ==> sqrt x
 warn  = sin x / cos x ==> tan x
 warn  = sinh x / cosh x ==> tanh x
 warn  = n `rem` 2 == 0 ==> even n
@@ -311,7 +392,6 @@ warn  = n `rem` 2 /= 0 ==> odd n
 warn  = not (even x) ==> odd x
 warn  = not (odd x) ==> even x
 warn  = x ** 0.5 ==> sqrt x
-warn  = x ^^ y ==> x ** y where _ = isLitInt y
 warn  "Use 1" = x ^ 0 ==> 1
 warn  = round (x - 0.5) ==> floor x
 
@@ -321,7 +401,6 @@ warn = mapM_ (writeChan a) ==> writeList2Chan a
 
 -- EXCEPTION
 
-error "Use Control.Exception.catch" = Prelude.catch ==> Control.Exception.catch where note = "Prelude.catch does not catch most exceptions"
 warn = flip Control.Exception.catch ==> handle
 warn = flip handle ==> Control.Exception.catch
 warn = flip (catchJust p) ==> handleJust p
@@ -334,12 +413,16 @@ error = a `seq` return a ==> Control.Exception.evaluate a
 error = toException NonTermination ==> nonTermination
 error = toException NestedAtomically ==> nestedAtomically
 
-
 -- WEAK POINTERS
 
 error = mkWeak a a b ==> mkWeakPtr a b
 error = mkWeak a (a, b) c ==> mkWeakPair a b c
 
+-- FOLDABLE
+
+error "Use Foldable.forM_" = (case m of Nothing -> return (); Just x -> f x) ==> Data.Foldable.forM_ m f
+error "Use Foldable.forM_" = when (isJust m) (f (fromJust m)) ==> Data.Foldable.forM_ m f
+
 -- EVALUATE
 
 -- TODO: These should be moved in to HSE\Evaluate.hs and applied
@@ -365,8 +448,8 @@ error "Evaluate" = foldr1 f [x] ==> x
 error "Evaluate" = scanr f z [] ==> [z]
 error "Evaluate" = scanr1 f [] ==> []
 error "Evaluate" = scanr1 f [x] ==> [x]
-error "Evaluate" = take n [] ==> []
-error "Evaluate" = drop n [] ==> []
+error "Evaluate" = take n [] ==> [] where note = IncreasesLaziness
+error "Evaluate" = drop n [] ==> [] where note = IncreasesLaziness
 error "Evaluate" = takeWhile p [] ==> []
 error "Evaluate" = dropWhile p [] ==> []
 error "Evaluate" = span p [] ==> ([],[])
@@ -378,18 +461,21 @@ error "Evaluate" = x / 1 ==> x
 error "Evaluate" = concat [a] ==> a
 error "Evaluate" = concat [] ==> []
 error "Evaluate" = zip [] [] ==> []
-error "Evaluate" = id x ==> x
 error "Evaluate" = const x y ==> x
 
 -- COMPLEX
 
-error "Use isPrefixOf" = (take i s == t) ==> _eval_ ((i == length t) && (t `Data.List.isPrefixOf` s))
-    where _ = (isList t || isLit t) && isLit i
+{-
+-- these would be a good idea, but we have not yet proven them and they seem to have side conditions
+error "Use isPrefixOf" = take (length t) s == t ==> t `Data.List.isPrefixOf` s
+error "Use isPrefixOf" = (take i s == t) ==> _eval_ ((i >= length t) && (t `Data.List.isPrefixOf` s))
+    where _ = (isList t || isLit t) && isPos i
+-}
 
 {-
 -- clever hint, but not actually a good idea
 warn  = (do a <- f; g a) ==> f >>= g
-    where _ = (isAtom f || isApp f) && notIn a g
+    where _ = (isAtom f || isApp f)
 -}
 
 test = hints named test are to allow people to put test code within hint files
@@ -411,10 +497,9 @@ yes = not (a == b) -- a /= b
 yes = not (a /= b) -- a == b
 yes = if a then 1 else if b then 1 else 2 -- if a || b then 1 else 2
 no  = if a then 1 else if b then 3 else 2
-yes = a >>= return . id -- Control.Monad.liftM id a
+yes = a >>= return . bob -- Control.Monad.liftM bob a
 yes = (x !! 0) + (x !! 2) -- head x
 yes = if b < 42 then [a] else [] -- [a | b < 42]
-yes = take 5 (foo xs) == "hello" -- "hello" `Data.List.isPrefixOf` foo xs
 no  = take n (foo xs) == "hello"
 yes = head (reverse xs) -- last xs
 yes = reverse xs `isPrefixOf` reverse ys -- isSuffixOf xs ys
@@ -462,11 +547,9 @@ yes = [v | v <- xs] -- xs
 no  = [Left x | Left x <- xs]
 yes = Map.union a b -- a `Map.union` b
 when p s = if p then s else return ()
-yes = x ^^ 18 -- x ** 18
 no = x ^^ 18.5
 instance Arrow (->) where first f = f *** id
 yes = fromInteger 12 -- 12
-yes = catch -- Control.Exception.catch
 import Prelude hiding (catch); no = catch
 import Control.Exception as E; no = E.catch
 main = do f; putStrLn $ show x -- print x
@@ -476,6 +559,31 @@ h x y = return $! x
 getInt = do { x <- readIO "0"; return $! (x :: Int) }
 foo = evaluate [12] -- return [12]
 test = \ a -> f a >>= \ b -> return (a, b)
+fooer input = catMaybes . map Just $ input -- mapMaybe Just
+main = print $ map (\_->5) [2,3,5] -- const 5
+main = x == a || x == b || x == c -- x `elem` [a,b,c]
+main = head $ drop n x
+main = head $ drop (-3) x -- x
+main = head $ drop 2 x -- x !! 2
+main = drop 0 x -- x
+main = take 0 x -- []
+main = take (-5) x -- []
+main = take (-y) x
+main = take 4 x
+main = let (first, rest) = (takeWhile p l, dropWhile p l) in rest -- span p l
+main = map $ \ d -> ([| $d |], [| $d |])
+pairs (x:xs) = map (\y -> (x,y)) xs ++ pairs xs
+{-# ANN foo "HLint: ignore" #-};foo = map f (map g x) -- @Ignore ???
+yes = fmap lines $ abc 123 -- lines Control.Applicative.<$> abc 123
+no = fmap lines $ abc $ def 123
+test = foo . not . not -- id
+test = map (not . not) xs -- id
+used = not . not . any (`notElem` special) . fst . derives -- any (`notElem` special) . fst . derives
+test = foo . id . map -- map
+test = food id xs
+yes = baz baz >> return () -- Control.Monad.void (baz baz)
+no = foo >>= bar >>= something >>= elsee >> return ()
+
 
 import Prelude \
 yes = flip mapM -- Control.Monad.forM
